"""Parse releases HTML and get releases."""

from lxml.html import HtmlElement


def parse_releases_html(html: HtmlElement) -> dict[str, list[tuple[str, str]]]:
    """Parse releases HTML.

    Args:
        html(HtmlElement): Target HTML
    Returns:
        dict: Key is the major version. Value is the list of tuples.
        A tuple consists of version and zip URL.
    """
    release_sections = html.cssselect(
        ".wp-block-wporg-release-tables__section"
    )
    ret: dict[str, list[tuple[str, str]]] = {}
    for release_section in release_sections:
        if release_section.attrib["id"] == "latest":
            continue
        elif release_section.attrib["id"] == "betas":
            section_title = "Beta & RC"
        elif release_section.attrib["id"] == "mu":
            section_title = "MU"
        elif release_section.attrib["id"].startswith("branch-"):
            section_title = (
                release_section.attrib["id"][7:8]
                + "."
                + release_section.attrib["id"][8:]
            )
        releases: list[tuple[str, str]] = []
        for row in release_section.cssselect("tbody tr"):
            release_version = row.cssselect("th")[0].text_content()
            zip_url = row.cssselect(
                ".wp-block-wporg-release-tables__cell-zip a"
            )[0].attrib["href"]
            releases.append((release_version, zip_url))
            ret[section_title] = releases
    return ret



"""Fetch and parse HTML module."""

import requests
from lxml.html import HtmlElement, fromstring


def fetch_html(url: str) -> HtmlElement:
    """Fetch HTML and parse.

    Args:
        url(str): Target URL

    Returns:
        HtmlElement: Parsed HTML
    """
    res = requests.get(url)
    res.raise_for_status()
    return fromstring(res.content)



"""Main program."""

import csv
import sys
import time
from argparse import ArgumentParser
from logging import INFO, basicConfig, getLogger

from requests.exceptions import HTTPError

from .html_getter import fetch_html
from .localize_checker import get_localized_release_pages
from .releases_parser import parse_releases_html


MAIN_RELEASES_URL = "https://wordpress.org/download/releases/"
BASE_SLEEP = 5

parser = ArgumentParser(description="WordPress list generator")
parser.add_argument("-o", "--output", required=True, help="Output filename")
args = parser.parse_args()

output_filename: str = args.output

logger = getLogger(__name__)
logger.setLevel(INFO)
basicConfig(stream=sys.stdout)

logger.info("Fetching: " + MAIN_RELEASES_URL)
base_html = fetch_html(MAIN_RELEASES_URL)
htmls = {MAIN_RELEASES_URL: base_html}
localized_pages = get_localized_release_pages(base_html)

rows: list[list[str]] = []

for locale, url in localized_pages.items():
    if url in htmls:
        locale_html = htmls[url]
    else:
        for i in range(3):
            time.sleep(BASE_SLEEP * (2 ** i))
            logger.info("Fetching: " + url)
            try:
                locale_html = fetch_html(url)
                break
            except HTTPError as e:
                if isinstance(e.args[0], str) and e.args[0].startswith("429"):
                    logger.warning("Failed to get " + url + ": " + str(i + 1))
                    continue
                else:
                    raise
        if "locale_html" not in locals():
            raise RuntimeError("Failed to get: " + url)
        htmls[url] = locale_html
    parse_results = parse_releases_html(locale_html)
    for branch, branch_data in parse_results.items():
        rows.extend(
            [locale, branch, item[0], item[1]]
            for item in branch_data
        )

with open(output_filename, "w", encoding="utf-8") as fp:
    writer = csv.writer(fp)
    writer.writerow(["Locale", "Branch", "Version", "URL"])
    writer.writerows(rows)



"""Get all localized WordPress release pages."""

from lxml.html import HtmlElement


def get_localized_release_pages(html: HtmlElement) -> dict[str, str]:
    """Get localized release page URLs.

    Args:
        html(HtmlElement): Page HTML

    Returns:
        dict: Language and release page URLs
    """
    ret: dict[str, str] = {}
    for elem in html.cssselect('link[rel="alternate"][hreflang]'):
        ret[elem.attrib["hreflang"]] = elem.attrib["href"]
    return ret



"""Test for html_getter module."""

import os
from concurrent.futures import ThreadPoolExecutor
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Self

from check_localized_wordpress_releases.html_getter import fetch_html


TEST_FILE_PATH = os.path.join(os.path.dirname(__file__), "assets", "en_releases.html")


class TestHTTPRequestHandler(BaseHTTPRequestHandler):
    """Test HTTP request handler for serving dummy HTML."""

    def do_GET(self: Self) -> None:
        """Handle GET request."""
        with open(TEST_FILE_PATH, "rb") as fp:
            content = fp.read()
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-Type", "text/html")
        self.send_header("Content-Length", str(len(content)))
        self.end_headers()
        self.wfile.write(content)

    def do_HEAD(self: Self) -> None:
        """Handle HEAD request."""
        with open(TEST_FILE_PATH, "rb") as fp:
            content = fp.read()
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-Type", "text/html")
        self.send_header("Content-Length", str(len(content)))
        self.end_headers()


def run_server(server: HTTPServer) -> None:
    """Run server.

    Args:
        server(HTTPServer): HTTP server
    """
    server.serve_forever()


def test_fetch_html() -> None:
    """Test fetch_html fuction."""
    server = HTTPServer(('127.0.0.1', 28080), TestHTTPRequestHandler)
    with ThreadPoolExecutor() as executor:
        executor.submit(run_server, server)
        res = fetch_html("http://127.0.0.1:28080/")
        try:
            assert res.attrib["lang"] == "en-US"
        finally:
            server.shutdown()



"""Test releases_parser module."""

import os

from lxml.html import fromstring

from check_localized_wordpress_releases.releases_parser import parse_releases_html


def test_parse_releases_html() -> None:
    """Test parse_releases_html function."""
    html_path = os.path.join(
        os.path.dirname(__file__), "assets", "en_releases.html"
    )
    with open(html_path, "r", encoding="utf-8") as fp:
        html = fromstring(fp.read())
    ret = parse_releases_html(html)
    assert ret["6.7"][0][0] == "6.7.2"
    assert ret["6.7"][0][1] == "https://wordpress.org/wordpress-6.7.2.zip"
    assert ret["Beta & RC"][1][0] == "6.8.1-RC1"
    assert ret["Beta & RC"][1][1] == "https://wordpress.org/wordpress-6.8.1-RC1.zip"



"""Test localize_checker module."""

import os

from lxml.html import fromstring

from check_localized_wordpress_releases.localize_checker import (
    get_localized_release_pages,
)


def test_get_localized_release_pages() -> None:
    """Test get_localized_release_pages function."""
    html_path = os.path.join(
        os.path.dirname(__file__), "assets", "en_releases.html"
    )
    with open(html_path, "r", encoding="utf-8") as fp:
        html = fromstring(fp.read())
    ret = get_localized_release_pages(html)
    assert ret["ja-jp"] == "https://ja.wordpress.org/download/releases/"
